# Generated by Django 4.2.2 on 2023-09-14 14:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0014_alter_annotationfield_validator_regex"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationfield",
            name="vocabulary",
            field=models.JSONField(blank=True, null=True),
        ),
    ]
