# Generated by Django 4.2.7 on 2023-12-14 07:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0019_alter_annotationpreset_modified_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationvalue",
            name="field",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="values",
                to="flow.annotationfield",
            ),
        ),
    ]
